const coral = @import("coral");

const handles = @import("./handles.zig");

const lina = @import("./lina.zig");

const sokol = @import("sokol");

const spirv = @import("./spirv.zig");

const std = @import("std");

pub const Effect = struct {
	shader: sokol.gfx.Shader,
	pipeline: sokol.gfx.Pipeline,
	has_properties_buffer: ?sokol.gfx.Buffer = null,

	pub fn deinit(self: *Effect) void {
		sokol.gfx.destroyPipeline(self.pipeline);
		sokol.gfx.destroyShader(self.shader);

		if (self.has_properties_buffer) |parameters_buffer| {
			sokol.gfx.destroyBuffer(parameters_buffer);
		}

		self.* = undefined;
	}

	pub fn init(desc: handles.Effect.Desc) !Effect {
		var spirv_unit = try spirv.Unit.init(coral.heap.allocator);

		defer spirv_unit.deinit();

		try spirv_unit.compile(&spirv.to_ops(@embedFile("./shaders/2d_default.vert.spv")), .vertex);
		try spirv_unit.compile(desc.fragment_spirv, .fragment);

		const shader = sokol.gfx.makeShader(spirv_unit.shader_desc());

		const pipeline = sokol.gfx.makePipeline(pipeline_desc: {
			var pipeline_desc = sokol.gfx.PipelineDesc{
				.label = "Effect pipeline",
				.layout = vertex_layout_state,
				.shader = shader,
				.index_type = .UINT16,
				.blend_color = .{.r = 1.0, .g = 1.0, .b = 1.0, .a = 1.0},
			};

			pipeline_desc.colors[0] = .{
				.write_mask = .RGBA,

				.blend = .{
					.enabled = true,
					.src_factor_rgb = .SRC_ALPHA,
					.dst_factor_rgb = .ONE_MINUS_SRC_ALPHA,
				},
			};

			break: pipeline_desc pipeline_desc;
		});

		errdefer {
			sokol.gfx.destroyPipeline(pipeline);
			sokol.gfx.destroyShader(shader);
		}

		if (spirv_unit.shader_stage(.fragment).has_storage_buffer[0]) |storage_buffer| {
			const properties_buffer = sokol.gfx.makeBuffer(.{
				.size = storage_buffer.minimum_size,
				.type = .STORAGEBUFFER,
				.usage = .STREAM,
			});

			errdefer {
				sokol.gfx.destroyBuffer(properties_buffer);
			}

			return .{
				.shader = shader,
				.pipeline = pipeline,
				.has_properties_buffer = properties_buffer,
			};
		} else {
			return .{
				.shader = shader,
				.pipeline = pipeline,
			};
		}
	}

	const vertex_layout_state = sokol.gfx.VertexLayoutState{
		.attrs = get: {
			var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

			attrs[0] = .{
				.format = .FLOAT2,
				.buffer_index = 0,
			};

			attrs[1] = .{
				.format = .FLOAT2,
				.buffer_index = 0,
			};

			attrs[2] = .{
				.format = .FLOAT2,
				.buffer_index = 1,
			};

			attrs[3] = .{
				.format = .FLOAT2,
				.buffer_index = 1,
			};

			attrs[4] = .{
				.format = .FLOAT2,
				.buffer_index = 1,
			};

			attrs[5] = .{
				.format = .UBYTE4N,
				.buffer_index = 1,
			};

			attrs[6] = .{
				.format = .FLOAT,
				.buffer_index = 1,
			};

			attrs[7] = .{
				.format = .FLOAT4,
				.buffer_index = 1,
			};

			break: get attrs;
		},

		.buffers = get: {
			var vertex_buffer_layouts = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

			vertex_buffer_layouts[1].step_func = .PER_INSTANCE;

			break: get vertex_buffer_layouts;
		},
	};
};

pub const Texture = union (enum) {
	render: Render,
	static: Static,

	const Render = struct {
		sampler: sokol.gfx.Sampler,
		color_image: sokol.gfx.Image,
		depth_image: sokol.gfx.Image,
		attachments: sokol.gfx.Attachments,
		view_buffer: sokol.gfx.Buffer,

		fn deinit(self: *Render) void {
			sokol.gfx.destroyImage(self.color_image);
			sokol.gfx.destroyImage(self.depth_image);
			sokol.gfx.destroySampler(self.sampler);
			sokol.gfx.destroyAttachments(self.attachments);
			sokol.gfx.destroyBuffer(self.view_buffer);

			self.* = undefined;
		}

		fn init(format: handles.Texture.Desc.Format, access: handles.Texture.Desc.Access.Render) std.mem.Allocator.Error!Render {
			const color_image = sokol.gfx.makeImage(.{
				.pixel_format = switch (format) {
					.rgba8 => sokol.gfx.PixelFormat.RGBA8,
					.bgra8 => sokol.gfx.PixelFormat.BGRA8,
				},

				.width = access.width,
				.height = access.height,
				.render_target = true,
			});

			const depth_image = sokol.gfx.makeImage(.{
				.width = access.width,
				.height = access.height,
				.render_target = true,
				.pixel_format = .DEPTH_STENCIL,
			});

			const attachments = sokol.gfx.makeAttachments(attachments_desc: {
				var desc = sokol.gfx.AttachmentsDesc{
					.depth_stencil = .{
						.image = depth_image,
					},
				};

				desc.colors[0] = .{
					.image = color_image,
				};

				break: attachments_desc desc;
			});

			const sampler = sokol.gfx.makeSampler(.{});

			const view_buffer = sokol.gfx.makeBuffer(.{
				.data = sokol.gfx.asRange(&lina.orthographic_projection(-1.0, 1.0, .{
					.left = 0,
					.top = 0,
					.right = @floatFromInt(access.width),
					.bottom = @floatFromInt(access.height),
				})),

				.type = .STORAGEBUFFER,
				.usage = .IMMUTABLE,
			});

			return .{
				.attachments = attachments,
				.sampler = sampler,
				.color_image = color_image,
				.depth_image = depth_image,
				.view_buffer = view_buffer,
			};
		}
	};

	const Static = struct {
		image: sokol.gfx.Image,
		sampler: sokol.gfx.Sampler,

		fn deinit(self: *Static) void {
			sokol.gfx.destroyImage(self.image);
			sokol.gfx.destroySampler(self.sampler);

			self.* = undefined;
		}

		fn init(format: handles.Texture.Desc.Format, access: handles.Texture.Desc.Access.Static) std.mem.Allocator.Error!Static {
			const image = sokol.gfx.makeImage(image_desc: {
				var desc = sokol.gfx.ImageDesc{
					.height = std.math.cast(i32, access.data.len / (access.width * format.byte_size())) orelse {
						return error.OutOfMemory;
					},

					.pixel_format = switch (format) {
						.rgba8 => sokol.gfx.PixelFormat.RGBA8,
						.bgra8 => sokol.gfx.PixelFormat.BGRA8,
					},

					.width = access.width,
				};

				desc.data.subimage[0][0] = sokol.gfx.asRange(access.data);

				break: image_desc desc;
			});

			const sampler = sokol.gfx.makeSampler(.{});

			errdefer {
				sokol.gfx.destroySampler(sampler);
				sokol.gfx.destroyImage(image);
			}

			return .{
				.image = image,
				.sampler = sampler,
			};
		}
	};

	pub fn deinit(self: *Texture) void {
		switch (self.*) {
			.static => |*static| static.deinit(),
			.render => |*render| render.deinit(),
		}
	}

	pub fn init(desc: handles.Texture.Desc) std.mem.Allocator.Error!Texture {
		return switch (desc.access) {
			.static => |static| .{.static = try Static.init(desc.format, static)},
			.render => |render| .{.render = try Render.init(desc.format, render)},
		};
	}
};
