const coral = @import("coral");

const ext = @cImport({
	@cInclude("spirv-cross/spirv_cross_c.h");
});

const sokol = @import("sokol");

const std = @import("std");

pub const Error = std.mem.Allocator.Error || error {
	UnsupportedTarget,
	InvalidSPIRV,
	UnsupportedSPIRV,
	UnsupportedBackend,
};

pub const Shader = struct {
	decompiled_source: [:0]const u8 = "",
	has_storage_buffer: [max_storage_buffers]?StorageBuffer = [_]?StorageBuffer{null} ** max_storage_buffers,
	has_image: [max_images]?Image = [_]?Image{null} ** max_images,
    has_sampler: [max_samplers]?Sampler = [_]?Sampler{null} ** max_samplers,
    has_image_sampler_pair: [max_image_sampler_pairs]?ImageSamplerPair = [_]?ImageSamplerPair{null} ** max_image_sampler_pairs,

	pub const Image = struct {
		is_multisampled: bool,
	};

	pub const ImageSamplerPair = struct {
		image_slot: usize,
		sampler_slot: usize,
		name: [:0]const u8,
	};

	pub const Sampler = enum {
		filtering,
		non_filtering,
		comparison,
	};

	pub const StorageBuffer = struct {
		minimum_size: usize,
		is_readonly: bool,
	};

	pub const max_images = 12;

	pub const max_image_sampler_pairs = 12;

	pub const max_samplers = 8;

	pub const max_storage_buffers = 8;

	fn reflect_image_samplers(self: *Shader, _: Stage, compiler: ext.spvc_compiler, resources: ext.spvc_resources) Error!void {
		var sampled_images: []const ext.spvc_reflected_resource = &.{};

		try to_error(ext.spvc_resources_get_resource_list_for_type(
			resources,
			ext.SPVC_RESOURCE_TYPE_SAMPLED_IMAGE,
			@ptrCast(&sampled_images.ptr),
			&sampled_images.len,
		));

		if (sampled_images.len > max_image_sampler_pairs) {
			return error.UnsupportedSPIRV;
		}

		for (0 .. sampled_images.len, sampled_images) |i, sampled_image| {
			const sampled_image_type = ext.spvc_compiler_get_type_handle(compiler, sampled_image.type_id);

			if (ext.spvc_type_get_basetype(sampled_image_type) != ext.SPVC_BASETYPE_SAMPLED_IMAGE) {
				return error.InvalidSPIRV;
			}

			try switch (ext.spvc_type_get_image_dimension(sampled_image_type)) {
				ext.SpvDim2D => {},
				else => error.InvalidSPIRV,
			};

			try switch (ext.spvc_type_get_basetype(ext.spvc_compiler_get_type_handle(compiler, ext.spvc_type_get_image_sampled_type(sampled_image_type)))) {
				ext.SPVC_BASETYPE_FP32 => {},
				else => error.InvalidSPIRV,
			};

			self.has_image[i] = .{
				.is_multisampled = ext.spvc_type_get_image_multisampled(sampled_image_type) != 0,
			};

			self.has_sampler[i] = .filtering;

			self.has_image_sampler_pair[i] = .{
				.name = std.mem.span(ext.spvc_compiler_get_name(compiler, sampled_image.id)),
				.image_slot = @intCast(i),
				.sampler_slot = @intCast(i),
			};
		}
	}

	fn reflect_storage_buffers(self: *Shader, stage: Stage, compiler: ext.spvc_compiler, resources: ext.spvc_resources) Error!void {
		var storage_buffers: []const ext.spvc_reflected_resource = &.{};

		try to_error(ext.spvc_resources_get_resource_list_for_type(
			resources,
			ext.SPVC_RESOURCE_TYPE_STORAGE_BUFFER,
			@ptrCast(&storage_buffers.ptr),
			&storage_buffers.len,
		));

		const binding_offset: c_uint = if (stage == .fragment) max_storage_buffers else 0;

		for (storage_buffers) |storage_buffer| {
			const binding = ext.spvc_compiler_get_decoration(compiler, storage_buffer.id, ext.SpvDecorationBinding);

			if (binding >= max_storage_buffers) {
				return error.InvalidSPIRV;
			}

			ext.spvc_compiler_set_decoration(compiler, storage_buffer.id, ext.SpvDecorationBinding, binding_offset + binding);

			var block_decorations: []const ext.SpvDecoration = &.{};

			try to_error(ext.spvc_compiler_get_buffer_block_decorations(
				compiler,
				storage_buffer.id,
				@ptrCast(&block_decorations.ptr),
				&block_decorations.len,
			));

			var minimum_size: usize = 0;

			try to_error(ext.spvc_compiler_get_declared_struct_size(compiler, ext.spvc_compiler_get_type_handle(compiler, storage_buffer.base_type_id), &minimum_size));

			self.has_storage_buffer[binding] = .{
				.is_readonly = std.mem.indexOfScalar(ext.SpvDecoration, block_decorations, ext.SpvDecorationNonWritable) != null,
				.minimum_size = minimum_size,
			};
		}
	}

	fn reflect_uniform_buffers(self: *Shader, _: Stage, compiler: ext.spvc_compiler, resources: ext.spvc_resources) Error!void {
		var uniform_buffers: []const ext.spvc_reflected_resource = &.{};

		try to_error(ext.spvc_resources_get_resource_list_for_type(
			resources,
			ext.SPVC_RESOURCE_TYPE_UNIFORM_BUFFER,
			@ptrCast(&uniform_buffers.ptr),
			&uniform_buffers.len,
		));

		if (uniform_buffers.len != 0) {
			return error.InvalidSPIRV;
		}

		// TODO: Support for older APIs?
		_ = self;
		_ = compiler;
	}

	pub fn stage_desc(self: Shader) sokol.gfx.ShaderStageDesc {
		var stage = sokol.gfx.ShaderStageDesc{
			.entry = "main",
			.source = self.decompiled_source,
		};

		for (0 .. max_storage_buffers) |slot| {
			const storage_buffer = &(self.has_storage_buffer[slot] orelse {
				continue;
			});

			stage.storage_buffers[slot] = .{
				.readonly = storage_buffer.is_readonly,
				.used = true,
			};
		}

		for (0 .. max_images) |slot| {
			const image = &(self.has_image[slot] orelse {
				continue;
			});

			stage.images[slot] = .{
				.multisampled = image.is_multisampled,
				.image_type = ._2D,
				.sample_type = .FLOAT,
				.used = true,
			};
		}

		for (0 .. max_samplers) |slot| {
			const sampler = &(self.has_sampler[slot] orelse {
				continue;
			});

			stage.samplers[slot] = .{
				.sampler_type = switch (sampler.*) {
					.filtering => .FILTERING,
					.non_filtering => .NONFILTERING,
					.comparison => .COMPARISON,
				},

				.used = true,
			};
		}

		for (0 .. max_image_sampler_pairs) |slot| {
			const image_sampler_pair = &(self.has_image_sampler_pair[slot] orelse {
				continue;
			});

			stage.image_sampler_pairs[slot] = .{
				.glsl_name = image_sampler_pair.name,
				.image_slot = @intCast(image_sampler_pair.image_slot),
				.sampler_slot = @intCast(image_sampler_pair.sampler_slot),
				.used = true,
			};
		}

		return stage;
	}
};

pub const Stage = enum {
	vertex,
	fragment,
};

pub const Unit = struct {
	arena: std.heap.ArenaAllocator,
	context: ext.spvc_context,
	attrs_used: u32 = 0,
	stages: [std.enums.values(Stage).len]Shader = .{.{}, .{}},

	pub fn compile(self: *Unit, spirv_ops: []const u32, stage: Stage) Error!void {
		const execution_model, const shader = switch (stage) {
			.vertex => .{ext.SpvExecutionModelVertex, &self.stages[0]},
			.fragment => .{ext.SpvExecutionModelFragment, &self.stages[1]},
		};

		const Backend = struct {
			target: ext.spvc_backend,
			option_values: []const struct {ext.spvc_compiler_option, c_uint},
		};

		const backend: Backend = try switch (sokol.gfx.queryBackend()) {
			.GLCORE => .{
				.target = ext.SPVC_BACKEND_GLSL,

				.option_values = &.{
					.{ext.SPVC_COMPILER_OPTION_GLSL_VERSION, 430},
					.{ext.SPVC_COMPILER_OPTION_GLSL_ES, @intFromBool(false)},
					.{ext.SPVC_COMPILER_OPTION_GLSL_VULKAN_SEMANTICS, @intFromBool(false)},
					.{ext.SPVC_COMPILER_OPTION_GLSL_EMIT_UNIFORM_BUFFER_AS_PLAIN_UNIFORMS, @intFromBool(true)},
					.{ext.SPVC_COMPILER_OPTION_FLIP_VERTEX_Y, @intFromBool(true)},
				},
			},

			else => error.UnsupportedBackend,
		};

		const compiler = parse_and_configure: {
			var parsed_ir: ext.spvc_parsed_ir = null;

			try to_error(ext.spvc_context_parse_spirv(self.context, spirv_ops.ptr, spirv_ops.len, &parsed_ir));

			var compiler: ext.spvc_compiler = null;

			try to_error(ext.spvc_context_create_compiler(self.context, backend.target, parsed_ir, ext.SPVC_CAPTURE_MODE_TAKE_OWNERSHIP, &compiler));
			try to_error(ext.spvc_compiler_build_combined_image_samplers(compiler));

			var combined_image_samplers: []const ext.spvc_combined_image_sampler = &.{};

			try to_error(ext.spvc_compiler_get_combined_image_samplers(compiler, @ptrCast(&combined_image_samplers.ptr), &combined_image_samplers.len));

			const arena_allocator = self.arena.allocator();
			var binding: u32 = 0;

			for (combined_image_samplers) |combined_image_sampler| {
				const name = try coral.utf8.alloc_formatted(arena_allocator, "{image_name}_{sampler_name}", .{
					.image_name = std.mem.span(ext.spvc_compiler_get_name(compiler, combined_image_sampler.image_id)),
					.sampler_name = std.mem.span(ext.spvc_compiler_get_name(compiler, combined_image_sampler.sampler_id)),
				});

				ext.spvc_compiler_set_name(compiler, combined_image_sampler.combined_id, name);
				ext.spvc_compiler_set_decoration(compiler, combined_image_sampler.combined_id, ext.SpvDecorationBinding, binding);

				binding += 1;
			}

			break: parse_and_configure compiler;
		};

		try to_error(ext.spvc_compiler_set_entry_point(compiler, "main", @intCast(execution_model)));

		const resources = create: {
			var resources: ext.spvc_resources = null;

			try to_error(ext.spvc_compiler_create_shader_resources(compiler, &resources));

			break: create resources;
		};

		try shader.reflect_uniform_buffers(stage, compiler, resources);
		try shader.reflect_storage_buffers(stage, compiler, resources);
		try shader.reflect_image_samplers(stage, compiler, resources);

		try to_error(ext.spvc_compiler_install_compiler_options(compiler, create: {
			var compiler_options: ext.spvc_compiler_options = null;

			try to_error(ext.spvc_compiler_create_compiler_options(compiler, &compiler_options));

			for (backend.option_values) |option_value| {
				const entry, const value = option_value;

				try to_error(ext.spvc_compiler_options_set_uint(compiler_options, entry, value));
			}

			break: create compiler_options;
		}));

		try to_error(ext.spvc_compiler_compile(compiler, @ptrCast(&shader.decompiled_source.ptr)));

		shader.decompiled_source.len = std.mem.span(shader.decompiled_source.ptr).len;

		std.log.info("{s}", .{shader.decompiled_source});
	}

	pub fn deinit(self: *Unit) void {
		ext.spvc_context_destroy(self.context);
		self.arena.deinit();

		self.* = undefined;
	}

	pub fn init(allocator: std.mem.Allocator) std.mem.Allocator.Error!Unit {
		var context: ext.spvc_context = null;

		if (ext.spvc_context_create(&context) != ext.SPVC_SUCCESS) {
			return error.OutOfMemory;
		}

		errdefer ext.spvc_context_destroy(context);

		ext.spvc_context_set_error_callback(context, log_context_errors, null);

		return .{
			.arena = std.heap.ArenaAllocator.init(allocator),
			.context = context,
		};
	}

	pub fn shader_desc(self: Unit) sokol.gfx.ShaderDesc {
		return .{
			.vs = self.shader_stage(.vertex).stage_desc(),
			.fs = self.shader_stage(.fragment).stage_desc(),
		};
	}

	pub fn shader_stage(self: *const Unit, stage: Stage) *const Shader {
		return &self.stages[@intFromEnum(stage)];
	}
};

fn log_context_errors(userdata: ?*anyopaque, error_message: [*c]const u8) callconv(.C) void {
	std.debug.assert(userdata == null);
	std.log.err("{s}", .{error_message});
}

fn to_error(result: ext.spvc_result) Error!void {
	return switch (result) {
		ext.SPVC_SUCCESS => {},
		ext.SPVC_ERROR_INVALID_SPIRV => error.InvalidSPIRV,
		ext.SPVC_ERROR_UNSUPPORTED_SPIRV => error.UnsupportedSPIRV,
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_INVALID_ARGUMENT, ext.SPVC_ERROR_INT_MAX => unreachable,
		else => unreachable,
	};
}

pub fn to_ops(raw: anytype) [raw.len / @alignOf(u32)]u32 {
	var ops: [raw.len / @alignOf(u32)]u32 = undefined;

	@memcpy(std.mem.sliceAsBytes(&ops), raw);

	return ops;
}
